import datetime

from dopynion.cards import Card, CardName

day1: set[CardName] = {
    CardName.COPPER,
    CardName.SILVER,
    CardName.GOLD,
    CardName.ESTATE,
    CardName.DUCHY,
    CardName.PROVINCE,
    CardName.FESTIVAL,
    CardName.LABORATORY,
    CardName.MARKET,
    CardName.SMITHY,
    CardName.VILLAGE,
    CardName.WOODCUTTER,
}

day2: set[CardName] = set()

day3: set[CardName] = {
    CardName.COUNCILROOM,
    CardName.CURSE,
    CardName.DISTANTSHORE,
    CardName.FARMINGVILLAGE,
    CardName.HIRELING,
    CardName.WITCH,
}

# begin hooks
day4: set[CardName] = {
    CardName.BANDIT,
    CardName.BUREAUCRAT,
    CardName.CHANCELLOR,
    CardName.GARDENS,
    CardName.MILITIA,
}

day5: set[CardName] = {
    CardName.ADVENTURER,
    CardName.CELLAR,
    CardName.FEAST,
    CardName.FORTUNETELLER,
    CardName.LIBRARY,
    CardName.MAGNATE,
    CardName.WORKSHOP,
}

day6: set[CardName] = {
    CardName.CHAPEL,
    CardName.MINE,
    CardName.MONEYLENDER,
    CardName.REMODEL,
    CardName.SWAP,  # TODO corriger le code de la carte, avec les nouveaux hooks
}

day7: set[CardName] = {
    CardName.ARTIFICER,
    CardName.COLONY,
    CardName.MARQUIS,
    CardName.POACHER,
    CardName.REMAKE,
}

day8: set[CardName] = {
    CardName.CURSEDGOLD,
    CardName.HARVEST,
    CardName.MAGPIE,
    CardName.PORT,
}

day9: set[CardName] = {
    CardName.FAIRGROUNDS,
    CardName.PLATINUM,
}


def init_day() -> None:
    allowed_cards = set()
    days = {
        datetime.date(2000, 1, 1): day1,
        datetime.date(2025, 10, 8): day2,
        datetime.date(2025, 10, 9): day3,
        datetime.date(2025, 10, 10): day4,
        datetime.date(2025, 10, 27): day5,
        datetime.date(2025, 10, 28): day6,
        datetime.date(2025, 11, 17): day7,
        datetime.date(2025, 11, 18): day8,
        datetime.date(2025, 11, 20): day9,
    }
    for day, cards in days.items():
        if day <= datetime.datetime.now(tz=datetime.UTC).date():
            allowed_cards |= cards

    for card_name in list(Card.types):
        if card_name not in allowed_cards:
            Card.types.pop(card_name)



from datetime import datetime
from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field


class CardName(StrEnum):  # Create with a metaclass
    ABANDONEDMINE = "abandonedmine"
    ACOLYTE = "acolyte"
    ACTINGTROUPE = "actingtroupe"
    ACTION_TOKEN = "action_token"
    ADVENTURER = "adventurer"
    ADVISOR = "advisor"
    ALCHEMIST = "alchemist"
    ALLY_ARCHITECTSGUILD = "ally_architectsguild"
    ALLY_BANDOFNOMADS = "ally_bandofnomads"
    ALLY_CAVEDWELLERS = "ally_cavedwellers"
    ALLY_CIRCLEOFWITCHES = "ally_circleofwitches"
    ALLY_CITYSTATE = "ally_citystate"
    ALLY_COASTALHAVEN = "ally_coastalhaven"
    ALLY_CRAFTERSGUILD = "ally_craftersguild"
    ALLY_DESERTGUIDES = "ally_desertguides"
    ALLY_FAMILYOFINVENTORS = "ally_familyofinventors"
    ALLY_FELLOWSHIPOFSCRIBES = "ally_fellowshipofscribes"
    ALLY_FORESTDWELLERS = "ally_forestdwellers"
    ALLY_GANGOFPICKPOCKETS = "ally_gangofpickpockets"
    ALLY_ISLANDFOLK = "ally_islandfolk"
    ALLY_LEAGUEOFBANKERS = "ally_leagueofbankers"
    ALLY_LEAGUEOFSHOPKEEPERS = "ally_leagueofshopkeepers"
    ALLY_MARKETTOWNS = "ally_markettowns"
    ALLY_MOUNTAINFOLK = "ally_mountainfolk"
    ALLY_ORDEROFASTROLOGERS = "ally_orderofastrologers"
    ALLY_ORDEROFMASONS = "ally_orderofmasons"
    ALLY_PEACEFULCULT = "ally_peacefulcult"
    ALLY_PLATEAUSHEPHERDS = "ally_plateaushepherds"
    ALLY_TRAPPERSLODGE = "ally_trapperslodge"
    ALLY_WOODWORKERSGUILD = "ally_woodworkersguild"
    ALTAR = "altar"
    AMBASSADOR = "ambassador"
    AMULET = "amulet"
    ANIMALFAIR = "animalfair"
    ANVIL = "anvil"
    APOTHECARY = "apothecary"
    APPRENTICE = "apprentice"
    ARCHER = "archer"
    ARCHIVE = "archive"
    ARMORY = "armory"
    ARTIFICER = "artificer"
    ARTISAN = "artisan"
    ASTROLABE = "astrolabe"
    AUGURS = "augurs"
    AVANTO = "avanto"
    BADOMENS = "badomens"
    BAGOFGOLD = "bagofgold"
    BAKER = "baker"
    BANDIT = "bandit"
    BANDITCAMP = "banditcamp"
    BANDOFMISFITS = "bandofmisfits"
    BANK = "bank"
    BARBARIAN = "barbarian"
    BARD = "bard"
    BARGE = "barge"
    BARON = "baron"
    BAT = "bat"
    BATTLEPLAN = "battleplan"
    BAUBLE = "bauble"
    BAZAAR = "bazaar"
    BEGGAR = "beggar"
    BERSERKER = "berserker"
    BISHOP = "bishop"
    BLACKCAT = "blackcat"
    BLACKMARKET = "blackmarket"
    BLACKSMITH = "blacksmith"
    BLESSEDVILLAGE = "blessedvillage"
    BLOCKADE = "blockade"
    BOON_THEEARTHSGIFT = "boon_theearthsgift"
    BOON_THEFIELDSGIFT = "boon_thefieldsgift"
    BOON_THEFLAMESGIFT = "boon_theflamesgift"
    BOON_THEFORESTSGIFT = "boon_theforestsgift"
    BOON_THEMOONSGIFT = "boon_themoonsgift"
    BOON_THEMOUNTAINSGIFT = "boon_themountainsgift"
    BOON_THERIVERSGIFT = "boon_theriversgift"
    BOON_THESEASGIFT = "boon_theseasgift"
    BOON_THESKYSGIFT = "boon_theskysgift"
    BOON_THESUNSGIFT = "boon_thesunsgift"
    BOON_THESWAMPSGIFT = "boon_theswampsgift"
    BOON_THEWINDSGIFT = "boon_thewindsgift"
    BORDERGUARD = "borderguard"
    BORDERVILLAGE = "bordervillage"
    BOUNTYHUNTER = "bountyhunter"
    BRIDGE = "bridge"
    BRIDGETROLL = "bridgetroll"
    BROKER = "broker"
    BUREAUCRAT = "bureaucrat"
    BUSTLINGVILLAGE = "bustlingvillage"
    BUTCHER = "butcher"
    BUY_TOKEN = "buy_token"
    CACHE = "cache"
    CAMELTRAIN = "cameltrain"
    CANDLESTICKMAKER = "candlestickmaker"
    CAPITAL = "capital"
    CAPITALCITY = "capitalcity"
    CAPTAIN = "captain"
    CARAVAN = "caravan"
    CARAVANGUARD = "caravanguard"
    CARD_TOKEN = "card_token"
    CARDINAL = "cardinal"
    CARGOSHIP = "cargoship"
    CARPENTER = "carpenter"
    CARTOGRAPHER = "cartographer"
    CASTLES = "castles"
    CATACOMBS = "catacombs"
    CATAPULT = "catapult"
    CATAPULTROCKS = "catapultrocks"
    CAULDRON = "cauldron"
    CAVALRY = "cavalry"
    CELLAR = "cellar"
    CEMETERY = "cemetery"
    CHAMPION = "champion"
    CHANCELLOR = "chancellor"
    CHANGELING = "changeling"
    CHAPEL = "chapel"
    CHARIOTRACE = "chariotrace"
    CHARLATAN = "charlatan"
    CHARM = "charm"
    CHURCH = "church"
    CITY = "city"
    CITYQUARTER = "cityquarter"
    CLASHES = "clashes"
    CLERK = "clerk"
    COBBLER = "cobbler"
    COFFERS_MAT = "coffers_mat"
    COFFERSVILLAGERS_MAT = "coffersvillagers_mat"
    COINOFTHEREALM = "coinoftherealm"
    COINSTOKENS = "coinstokens"
    COLLECTION = "collection"
    COLONY = "colony"
    COLONY_2ND = "colony_2nd"
    CONCLAVE = "conclave"
    CONJURER = "conjurer"
    CONSPIRATOR = "conspirator"
    CONTRABAND = "contraband"
    CONTRACT = "contract"
    COPPER = "copper"
    COPPER2 = "copper2"
    COPPER2_2ND = "copper2_2nd"
    COPPER2ND = "copper2nd"
    COPPERSMITH = "coppersmith"
    CORSAIR = "corsair"
    COUNCILROOM = "councilroom"
    COUNT = "count"
    COUNTERFEIT = "counterfeit"
    COUNTINGHOUSE = "countinghouse"
    COURIER = "courier"
    COURTIER = "courtier"
    COURTYARD = "courtyard"
    COVEN = "coven"
    CROSSROADS = "crossroads"
    CROWN = "crown"
    CRUMBLINGCASTLE = "crumblingcastle"
    CRYPT = "crypt"
    CRYSTALBALL = "crystalball"
    CULTIST = "cultist"
    CURSE = "curse"
    CURSE2 = "curse2"
    CURSE2_2ND = "curse2_2nd"
    CURSE2ND = "curse2nd"
    CURSEDGOLD = "cursedgold"
    CURSEDVILLAGE = "cursedvillage"
    CUTPURSE = "cutpurse"
    DAMEANNA = "dameanna"
    DAMEJOSEPHINE = "damejosephine"
    DAMEMOLLY = "damemolly"
    DAMENATALIE = "damenatalie"
    DAMESYLVIA = "damesylvia"
    DEATHCART = "deathcart"
    DEBTTOKENS = "debttokens"
    DELUDED = "deluded"
    DELUSION = "delusion"
    DENOFSIN = "denofsin"
    DESTRIER = "destrier"
    DEVELOP = "develop"
    DEVILSWORKSHOP = "devilsworkshop"
    DIADEM = "diadem"
    DIPLOMAT = "diplomat"
    DISCIPLE = "disciple"
    DISMANTLE = "dismantle"
    DISPLACE = "displace"
    DISTANTLANDS = "distantlands"
    DISTANTSHORE = "distantshore"
    DOCTOR = "doctor"
    DRUID = "druid"
    DUCAT = "ducat"
    DUCHESS = "duchess"
    DUCHY = "duchy"
    DUCHY2 = "duchy2"
    DUCHY2_2ND = "duchy2_2nd"
    DUCHY2ND = "duchy2nd"
    DUKE = "duke"
    DUNGEON = "dungeon"
    DUPLICATE = "duplicate"
    ELDER = "elder"
    EMBARGO = "embargo"
    EMBARGOTOKENS = "embargotokens"
    EMBASSY = "embassy"
    EMISSARY = "emissary"
    EMPORIUM = "emporium"
    ENCAMPMENT = "encampment"
    ENCAMPMENTPLUNDER = "encampmentplunder"
    ENCHANTRESS = "enchantress"
    ENGINEER = "engineer"
    ENVIOUS = "envious"
    ENVOY = "envoy"
    ENVY = "envy"
    ESTATE = "estate"
    ESTATE2 = "estate2"
    ESTATE2_2ND = "estate2_2nd"
    ESTATE2ND = "estate2nd"
    ESTATE_TOKEN = "estate_token"
    EVENT_ADVANCE = "event_advance"
    EVENT_ALLIANCE = "event_alliance"
    EVENT_ALMS = "event_alms"
    EVENT_ANNEX = "event_annex"
    EVENT_BALL = "event_ball"
    EVENT_BANISH = "event_banish"
    EVENT_BANQUET = "event_banquet"
    EVENT_BARGAIN = "event_bargain"
    EVENT_BONFIRE = "event_bonfire"
    EVENT_BORROW = "event_borrow"
    EVENT_COMMERCE = "event_commerce"
    EVENT_CONQUEST = "event_conquest"
    EVENT_DELAY = "event_delay"
    EVENT_DELVE = "event_delve"
    EVENT_DEMAND = "event_demand"
    EVENT_DESPERATION = "event_desperation"
    EVENT_DOMINATE = "event_dominate"
    EVENT_DONATE = "event_donate"
    EVENT_ENCLAVE = "event_enclave"
    EVENT_ENHANCE = "event_enhance"
    EVENT_EXPEDITION = "event_expedition"
    EVENT_FERRY = "event_ferry"
    EVENT_GAMBLE = "event_gamble"
    EVENT_INHERITANCE = "event_inheritance"
    EVENT_INVEST = "event_invest"
    EVENT_LOSTARTS = "event_lostarts"
    EVENT_MARCH = "event_march"
    EVENT_MISSION = "event_mission"
    EVENT_PATHFINDING = "event_pathfinding"
    EVENT_PILGRIMAGE = "event_pilgrimage"
    EVENT_PLAN = "event_plan"
    EVENT_POPULATE = "event_populate"
    EVENT_PURSUE = "event_pursue"
    EVENT_QUEST = "event_quest"
    EVENT_RAID = "event_raid"
    EVENT_REAP = "event_reap"
    EVENT_RIDE = "event_ride"
    EVENT_RITUAL = "event_ritual"
    EVENT_SALTTHEEARTH = "event_salttheearth"
    EVENT_SAVE = "event_save"
    EVENT_SCOUTINGPARTY = "event_scoutingparty"
    EVENT_SEAWAY = "event_seaway"
    EVENT_SEIZETHEDAY = "event_seizetheday"
    EVENT_STAMPEDE = "event_stampede"
    EVENT_SUMMON = "event_summon"
    EVENT_TAX = "event_tax"
    EVENT_TOIL = "event_toil"
    EVENT_TRADE = "event_trade"
    EVENT_TRAINING = "event_training"
    EVENT_TRANSPORT = "event_transport"
    EVENT_TRAVELLINGFAIR = "event_travellingfair"
    EVENT_TRIUMPH = "event_triumph"
    EVENT_WEDDING = "event_wedding"
    EVENT_WINDFALL = "event_windfall"
    EXILE_MAT = "exile_mat"
    EXORCIST = "exorcist"
    EXPAND = "expand"
    EXPERIMENT = "experiment"
    EXPLORER = "explorer"
    FAIRGROUNDS = "fairgrounds"
    FAITHFULHOUND = "faithfulhound"
    FALCONER = "falconer"
    FAMILIAR = "familiar"
    FAMINE = "famine"
    FARMERSMARKET = "farmersmarket"
    FARMINGVILLAGE = "farmingvillage"
    FARMLAND = "farmland"
    FEAR = "fear"
    FEAST = "feast"
    FEODUM = "feodum"
    FESTIVAL = "festival"
    FISHERMAN = "fisherman"
    FISHINGVILLAGE = "fishingvillage"
    FLAG = "flag"
    FLAGBEARER = "flagbearer"
    FOLLOWERS = "followers"
    FOOL = "fool"
    FOOLSGOLD = "foolsgold"
    FORAGER = "forager"
    FORGE = "forge"
    FORTRESS = "fortress"
    FORTS = "forts"
    FORTUNE = "fortune"
    FORTUNETELLER = "fortuneteller"
    FORUM = "forum"
    FUGITIVE = "fugitive"
    GALLERIA = "galleria"
    GARDENS = "gardens"
    GARRISON = "garrison"
    GATEKEEPER = "gatekeeper"
    GEAR = "gear"
    GHOST = "ghost"
    GHOSTSHIP = "ghostship"
    GHOSTTOWN = "ghosttown"
    GIANT = "giant"
    GLADIATOR = "gladiator"
    GLADIATORFORTUNE = "gladiatorfortune"
    GOAT = "goat"
    GOATHERD = "goatherd"
    GOLD = "gold"
    GOLD2 = "gold2"
    GOLD2_2ND = "gold2_2nd"
    GOLD2ND = "gold2nd"
    GOLEM = "golem"
    GOONS = "goons"
    GOVERNOR = "governor"
    GRANDCASTLE = "grandcastle"
    GRANDMARKET = "grandmarket"
    GRAVEROBBER = "graverobber"
    GREATHALL = "greathall"
    GREED = "greed"
    GROOM = "groom"
    GROUNDSKEEPER = "groundskeeper"
    GUARDDOG = "guarddog"
    GUARDIAN = "guardian"
    GUIDE = "guide"
    GUILDMASTER = "guildmaster"
    HAGGLER = "haggler"
    HAMLET = "hamlet"
    HARBINGER = "harbinger"
    HAREM = "harem"
    HARVEST = "harvest"
    HAUNTEDCASTLE = "hauntedcastle"
    HAUNTEDMIRROR = "hauntedmirror"
    HAUNTEDWOODS = "hauntedwoods"
    HAUNTING = "haunting"
    HAVEN = "haven"
    HERALD = "herald"
    HERBALIST = "herbalist"
    HERBGATHERER = "herbgatherer"
    HERMIT = "hermit"
    HERO = "hero"
    HIDEOUT = "hideout"
    HIGHWAY = "highway"
    HIGHWAYMAN = "highwayman"
    HILLFORT = "hillfort"
    HIRELING = "hireling"
    HOARD = "hoard"
    HORN = "horn"
    HORNOFPLENTY = "hornofplenty"
    HORSE = "horse"
    HORSETRADERS = "horsetraders"
    HOSTELRY = "hostelry"
    HOVEL = "hovel"
    HUMBLECASTLE = "humblecastle"
    HUNTER = "hunter"
    HUNTINGGROUNDS = "huntinggrounds"
    HUNTINGLODGE = "huntinglodge"
    HUNTINGPARTY = "huntingparty"
    IDOL = "idol"
    ILLGOTTENGAINS = "illgottengains"
    IMP = "imp"
    IMPORTER = "importer"
    IMPROVE = "improve"
    INN = "inn"
    INNKEEPER = "innkeeper"
    INVENTOR = "inventor"
    INVESTMENT = "investment"
    IRONMONGER = "ironmonger"
    IRONWORKS = "ironworks"
    ISLAND = "island"
    ISLAND_MAT = "island_mat"
    JACKOFALLTRADES = "jackofalltrades"
    JESTER = "jester"
    JOURNEY_TOKEN = "journey_token"
    JOURNEYMAN = "journeyman"
    JUNKDEALER = "junkdealer"
    KEY = "key"
    KILN = "kiln"
    KINGSCASTLE = "kingscastle"
    KINGSCOURT = "kingscourt"
    KNIGHTS = "knights"
    LABORATORY = "laboratory"
    LACKEYS = "lackeys"
    LANDMARK_AQUEDUCT = "landmark_aqueduct"
    LANDMARK_ARENA = "landmark_arena"
    LANDMARK_BANDITFORT = "landmark_banditfort"
    LANDMARK_BASILICA = "landmark_basilica"
    LANDMARK_BATHS = "landmark_baths"
    LANDMARK_BATTLEFIELD = "landmark_battlefield"
    LANDMARK_COLONNADE = "landmark_colonnade"
    LANDMARK_DEFILEDSHRINE = "landmark_defiledshrine"
    LANDMARK_FOUNTAIN = "landmark_fountain"
    LANDMARK_KEEP = "landmark_keep"
    LANDMARK_LABYRINTH = "landmark_labyrinth"
    LANDMARK_MOUNTAINPASS = "landmark_mountainpass"
    LANDMARK_MUSEUM = "landmark_museum"
    LANDMARK_OBELISK = "landmark_obelisk"
    LANDMARK_ORCHARD = "landmark_orchard"
    LANDMARK_PALACE = "landmark_palace"
    LANDMARK_TOMB = "landmark_tomb"
    LANDMARK_TOWER = "landmark_tower"
    LANDMARK_TRIUMPHALARCH = "landmark_triumphalarch"
    LANDMARK_WALL = "landmark_wall"
    LANDMARK_WOLFDEN = "landmark_wolfden"
    LANTERN = "lantern"
    LEGIONARY = "legionary"
    LEPRECHAUN = "leprechaun"
    LIBRARY = "library"
    LICH = "lich"
    LIGHTHOUSE = "lighthouse"
    LIVERY = "livery"
    LOAN = "loan"
    LOCUSTS = "locusts"
    LOOKOUT = "lookout"
    LOSTCITY = "lostcity"
    LOSTINTHEWOODS = "lostinthewoods"
    LUCKYCOIN = "luckycoin"
    LURKER = "lurker"
    MADMAN = "madman"
    MAGICLAMP = "magiclamp"
    MAGNATE = "magnate"
    MAGPIE = "magpie"
    MANDARIN = "mandarin"
    MARAUDER = "marauder"
    MARGRAVE = "margrave"
    MARKET = "market"
    MARKETSQUARE = "marketsquare"
    MARQUIS = "marquis"
    MASQUERADE = "masquerade"
    MASTERMIND = "mastermind"
    MASTERPIECE = "masterpiece"
    MENAGERIE = "menagerie"
    MERCENARY = "mercenary"
    MERCHANT = "merchant"
    MERCHANTCAMP = "merchantcamp"
    MERCHANTGUILD = "merchantguild"
    MERCHANTSHIP = "merchantship"
    MESSENGER = "messenger"
    MILITIA = "militia"
    MILL = "mill"
    MILLER = "miller"
    MINE = "mine"
    MININGVILLAGE = "miningvillage"
    MINION = "minion"
    MINT = "mint"
    MINUSCARD_TOKEN = "minuscard_token"
    MINUSCOIN_TOKEN = "minuscoin_token"
    MINUSCOST_TOKEN = "minuscost_token"
    MISER = "miser"
    MISERABLE = "miserable"
    MISERY = "misery"
    MOAT = "moat"
    MODIFY = "modify"
    MONASTERY = "monastery"
    MONEYLENDER = "moneylender"
    MONKEY = "monkey"
    MONUMENT = "monument"
    MOUNTAINVILLAGE = "mountainvillage"
    MOUNTEBANK = "mountebank"
    MYSTIC = "mystic"
    NATIVE_VILLAGE_MAT = "native_village_mat"
    NATIVEVILLAGE = "nativevillage"
    NAVIGATOR = "navigator"
    NECROMANCER = "necromancer"
    NECROPOLIS = "necropolis"
    NIGHTWATCHMAN = "nightwatchman"
    NOBLEBRIGAND = "noblebrigand"
    NOBLES = "nobles"
    NOMADCAMP = "nomadcamp"
    NOMADS = "nomads"
    OASIS = "oasis"
    ODYSSEYS = "odysseys"
    OLDMAP = "oldmap"
    OLDWITCH = "oldwitch"
    OPULENTCASTLE = "opulentcastle"
    ORACLE = "oracle"
    OUTPOST = "outpost"
    OVERGROWNESTATE = "overgrownestate"
    OVERLORD = "overlord"
    PADDOCK = "paddock"
    PAGE = "page"
    PASTURE = "pasture"
    PATRICIAN = "patrician"
    PATRICIANEMPORIUM = "patricianemporium"
    PATROL = "patrol"
    PATRON = "patron"
    PAWN = "pawn"
    PEARLDIVER = "pearldiver"
    PEASANT = "peasant"
    PEDDLER = "peddler"
    PHILOSOPHERSSTONE = "philosophersstone"
    PILLAGE = "pillage"
    PIRATE = "pirate"
    PIRATE_SHIP_MAT = "pirate_ship_mat"
    PIRATESHIP = "pirateship"
    PIXIE = "pixie"
    PLAGUE = "plague"
    PLATINUM = "platinum"
    PLATINUM_2ND = "platinum_2nd"
    PLAZA = "plaza"
    PLUNDER = "plunder"
    PLUSCOIN_TOKEN = "pluscoin_token"
    POACHER = "poacher"
    POOKA = "pooka"
    POORHOUSE = "poorhouse"
    PORT = "port"
    POSSESSION = "possession"
    POTION = "potion"
    POTION_2ND = "potion_2nd"
    POUCH = "pouch"
    POVERTY = "poverty"
    PRIEST = "priest"
    PRINCE = "prince"
    PRINCESS = "princess"
    PROCESSION = "procession"
    PROJECT_ACADEMY = "project_academy"
    PROJECT_BARRACKS = "project_barracks"
    PROJECT_CANAL = "project_canal"
    PROJECT_CAPITALISM = "project_capitalism"
    PROJECT_CATHEDRAL = "project_cathedral"
    PROJECT_CITADEL = "project_citadel"
    PROJECT_CITYGATE = "project_citygate"
    PROJECT_CROPROTATION = "project_croprotation"
    PROJECT_EXPLORATION = "project_exploration"
    PROJECT_FAIR = "project_fair"
    PROJECT_FLEET = "project_fleet"
    PROJECT_GUILDHALL = "project_guildhall"
    PROJECT_INNOVATION = "project_innovation"
    PROJECT_PAGEANT = "project_pageant"
    PROJECT_PIAZZA = "project_piazza"
    PROJECT_ROADNETWORK = "project_roadnetwork"
    PROJECT_SEWERS = "project_sewers"
    PROJECT_SILOS = "project_silos"
    PROJECT_SINISTERPLOT = "project_sinisterplot"
    PROJECT_STARCHART = "project_starchart"
    PROVINCE = "province"
    PROVINCE2 = "province2"
    PROVINCE2_2ND = "province2_2nd"
    PROVINCE2ND = "province2nd"
    QUARRY = "quarry"
    RABBLE = "rabble"
    RAIDER = "raider"
    RANGER = "ranger"
    RATCATCHER = "ratcatcher"
    RATS = "rats"
    RAZE = "raze"
    REBUILD = "rebuild"
    RECRUITER = "recruiter"
    RELIC = "relic"
    REMAKE = "remake"
    REMODEL = "remodel"
    REPLACE = "replace"
    RESEARCH = "research"
    ROCKS = "rocks"
    ROGUE = "rogue"
    ROYALBLACKSMITH = "royalblacksmith"
    ROYALCARRIAGE = "royalcarriage"
    ROYALGALLEY = "royalgalley"
    ROYALSEAL = "royalseal"
    RUINEDLIBRARY = "ruinedlibrary"
    RUINEDMARKET = "ruinedmarket"
    RUINEDVILLAGE = "ruinedvillage"
    SABOTEUR = "saboteur"
    SACREDGROVE = "sacredgrove"
    SACRIFICE = "sacrifice"
    SAGE = "sage"
    SAILOR = "sailor"
    SALVAGER = "salvager"
    SANCTUARY = "sanctuary"
    SAUNA = "sauna"
    SAUNAAVANTO = "saunaavanto"
    SCAVENGER = "scavenger"
    SCEPTER = "scepter"
    SCHEME = "scheme"
    SCHOLAR = "scholar"
    SCOUT = "scout"
    SCRAP = "scrap"
    SCRYINGPOOL = "scryingpool"
    SCULPTOR = "sculptor"
    SEACHART = "seachart"
    SEAHAG = "seahag"
    SEAWITCH = "seawitch"
    SECRETCAVE = "secretcave"
    SECRETCHAMBER = "secretchamber"
    SECRETPASSAGE = "secretpassage"
    SEER = "seer"
    SENTINEL = "sentinel"
    SENTRY = "sentry"
    SETTLERS = "settlers"
    SETTLERSBUSTLINGVILLAGE = "settlersbustlingvillage"
    SHANTYTOWN = "shantytown"
    SHEEPDOG = "sheepdog"
    SHEPHERD = "shepherd"
    SIBYL = "sibyl"
    SILKMERCHANT = "silkmerchant"
    SILKROAD = "silkroad"
    SILVER = "silver"
    SILVER2 = "silver2"
    SILVER2_2ND = "silver2_2nd"
    SILVER2ND = "silver2nd"
    SIRBAILEY = "sirbailey"
    SIRDESTRY = "sirdestry"
    SIRMARTIN = "sirmartin"
    SIRMICHAEL = "sirmichael"
    SIRVANDER = "sirvander"
    SKIRMISHER = "skirmisher"
    SKULK = "skulk"
    SLEIGH = "sleigh"
    SMALLCASTLE = "smallcastle"
    SMITHY = "smithy"
    SMUGGLERS = "smugglers"
    SNOWYVILLAGE = "snowyvillage"
    SOLDIER = "soldier"
    SOOTHSAYER = "soothsayer"
    SORCERER = "sorcerer"
    SORCERESS = "sorceress"
    SOUK = "souk"
    SPECIALIST = "specialist"
    SPICEMERCHANT = "spicemerchant"
    SPICES = "spices"
    SPOILS = "spoils"
    SPRAWLINGCASTLE = "sprawlingcastle"
    SPY = "spy"
    SQUIRE = "squire"
    STABLES = "stables"
    STASH = "stash"
    STEWARD = "steward"
    STOCKPILE = "stockpile"
    STONEMASON = "stonemason"
    STOREROOM = "storeroom"
    STORYTELLER = "storyteller"
    STRONGHOLD = "stronghold"
    STUDENT = "student"
    SUNKENTREASURE = "sunkentreasure"
    SUPPLIES = "supplies"
    SURVIVORS = "survivors"
    SWAMPHAG = "swamphag"
    SWAP = "swap"
    SWASHBUCKLER = "swashbuckler"
    SWINDLER = "swindler"
    SYCOPHANT = "sycophant"
    TACTICIAN = "tactician"
    TALISMAN = "talisman"
    TAVERN_MAT = "tavern_mat"
    TAXMAN = "taxman"
    TEACHER = "teacher"
    TEMPLE = "temple"
    TENT = "tent"
    TERRITORY = "territory"
    THIEF = "thief"
    THRONEROOM = "throneroom"
    TIARA = "tiara"
    TIDEPOOLS = "tidepools"
    TORMENTOR = "tormentor"
    TORTURER = "torturer"
    TOURNAMENT = "tournament"
    TOWN = "town"
    TOWNCRIER = "towncrier"
    TOWNSFOLK = "townsfolk"
    TRACKER = "tracker"
    TRADE_ROUTE_MAT = "trade_route_mat"
    TRADER = "trader"
    TRADEROUTE = "traderoute"
    TRADINGPOST = "tradingpost"
    TRAGICHERO = "tragichero"
    TRAIL = "trail"
    TRANSMOGRIFY = "transmogrify"
    TRANSMUTE = "transmute"
    TRASHING_TOKEN = "trashing_token"
    TREASURECHEST = "treasurechest"
    TREASUREHUNTER = "treasurehunter"
    TREASUREMAP = "treasuremap"
    TREASURER = "treasurer"
    TREASURETROVE = "treasuretrove"
    TREASURY = "treasury"
    TRIBUTE = "tribute"
    TRUSTYSTEED = "trustysteed"
    TUNNEL = "tunnel"
    TWICEMISERABLE = "twicemiserable"
    UNDERLING = "underling"
    UNIVERSITY = "university"
    UPGRADE = "upgrade"
    URCHIN = "urchin"
    VAGRANT = "vagrant"
    VAMPIRE = "vampire"
    VASSAL = "vassal"
    VAULT = "vault"
    VENTURE = "venture"
    VILLA = "villa"
    VILLAGE = "village"
    VILLAGEGREEN = "villagegreen"
    VILLAIN = "villain"
    VINEYARD = "vineyard"
    VOYAGE = "voyage"
    VP_MAT_1 = "vp_mat_1"
    VP_MAT_2 = "vp_mat_2"
    VP_MAT_3 = "vp_mat_3"
    VP_MAT_4 = "vp_mat_4"
    VP_MAT_5 = "vp_mat_5"
    VP_MAT_6 = "vp_mat_6"
    VP_MAT_7 = "vp_mat_7"
    VP_MAT_8 = "vp_mat_8"
    VPTOKENS = "vptokens"
    WALLEDVILLAGE = "walledvillage"
    WANDERINGMINSTREL = "wanderingminstrel"
    WAR = "war"
    WARCHEST = "warchest"
    WAREHOUSE = "warehouse"
    WARLORD = "warlord"
    WARRIOR = "warrior"
    WATCHTOWER = "watchtower"
    WAY_WAYOFTHEBUTTERFLY = "way_wayofthebutterfly"
    WAY_WAYOFTHECAMEL = "way_wayofthecamel"
    WAY_WAYOFTHECHAMELEON = "way_wayofthechameleon"
    WAY_WAYOFTHEFROG = "way_wayofthefrog"
    WAY_WAYOFTHEGOAT = "way_wayofthegoat"
    WAY_WAYOFTHEHORSE = "way_wayofthehorse"
    WAY_WAYOFTHEMOLE = "way_wayofthemole"
    WAY_WAYOFTHEMONKEY = "way_wayofthemonkey"
    WAY_WAYOFTHEMOUSE = "way_wayofthemouse"
    WAY_WAYOFTHEMULE = "way_wayofthemule"
    WAY_WAYOFTHEOTTER = "way_wayoftheotter"
    WAY_WAYOFTHEOWL = "way_wayoftheowl"
    WAY_WAYOFTHEOX = "way_wayoftheox"
    WAY_WAYOFTHEPIG = "way_wayofthepig"
    WAY_WAYOFTHERAT = "way_wayoftherat"
    WAY_WAYOFTHESEAL = "way_wayoftheseal"
    WAY_WAYOFTHESHEEP = "way_wayofthesheep"
    WAY_WAYOFTHESQUIRREL = "way_wayofthesquirrel"
    WAY_WAYOFTHETURTLE = "way_wayoftheturtle"
    WAY_WAYOFTHEWORM = "way_wayoftheworm"
    WAYFARER = "wayfarer"
    WEAVER = "weaver"
    WEREWOLF = "werewolf"
    WHARF = "wharf"
    WHEELWRIGHT = "wheelwright"
    WILDHUNT = "wildhunt"
    WILLOWISP = "willowisp"
    WINEMERCHANT = "winemerchant"
    WISH = "wish"
    WISHINGWELL = "wishingwell"
    WITCH = "witch"
    WITCHSHUT = "witchshut"
    WIZARDS = "wizards"
    WOODCUTTER = "woodcutter"
    WORKERSVILLAGE = "workersvillage"
    WORKSHOP = "workshop"
    YOUNGWITCH = "youngwitch"
    ZOMBIEAPPRENTICE = "zombieapprentice"
    ZOMBIEMASON = "zombiemason"
    ZOMBIESPY = "zombiespy"
    NONE = "NONE"

    def __repr__(self) -> str:
        return self.value.title()


class Cards(BaseModel):
    quantities: dict[CardName, int] = Field(default_factory=dict)


class Player(BaseModel):
    name: str
    hand: Cards | None
    score: int


class Game(BaseModel):
    finished: bool
    players: list[Player]
    stock: Cards


class ActionRecord(BaseModel):
    action: str
    player: Player
    score: int


class ErrorRecord(BaseModel):
    error: str
    player: Player
    type: Literal["error", "warning"] = "error"


class CardNameAndHand(BaseModel):
    card_name: CardName
    hand: list[CardName]


class Hand(BaseModel):
    hand: list[CardName]


class PossibleCards(BaseModel):
    possible_cards: list[CardName]


class MoneyCardsInHand(BaseModel):
    money_in_hand: list[CardName]


HookCallArgs = CardNameAndHand | Hand | PossibleCards | MoneyCardsInHand | None
HookCallResult = CardName | bool | None


class HookCallRecord(BaseModel):
    name: str
    player: Player
    args: HookCallArgs


class HookResultRecord(BaseModel):
    player: Player
    result: HookCallResult


class PlayerTurnRecord(BaseModel):
    actions: list[ActionRecord | ErrorRecord | HookCallRecord | HookResultRecord] = (
        Field(default_factory=list)
    )


class GameRecord(BaseModel):
    date: datetime
    stock: Cards
    turns: list[PlayerTurnRecord] = Field(default_factory=list)
    scores: dict[str, int] = Field(default_factory=dict)



from __future__ import annotations

import inspect
import logging
import random
import sys
from collections import defaultdict
from typing import TYPE_CHECKING, ClassVar, Final

from dopynion.data_model import (
    CardName,
    CardNameAndHand,
    Cards,
    Hand,
    MoneyCardsInHand,
    PossibleCards,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from types import TracebackType

    from dopynion.player import Player

logger = logging.getLogger(__name__)


# TODO ajout robustesse sur action invalide


class ErrorManager:
    def __init__(self, player: Player) -> None:
        self.player = player

    def __enter__(self) -> None:
        pass

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        if exc_type is not None:
            self.player.game.record.add_error(
                f"Error: {exc_type!r} {exc_val!r}",
                self.player,
            )
            self.player.eliminate()
            self.player.game.save()
        return True


class ClassNameRepr(type):
    def __repr__(cls) -> str:
        return cls.__name__


class Card(metaclass=ClassNameRepr):
    types: ClassVar[dict[CardName, type[Card]]] = {}
    card_set = "baseset"
    name = "Unknown"
    cost = 10_000
    money = 0
    is_action = False
    is_kingdom = True
    is_treasure = False
    is_victory = False
    more_cards_from_deck = 0
    more_purchases = 0
    more_actions = 0
    more_money = 0
    victory_points = 0

    @classmethod
    def card_name(cls) -> CardName:
        return CardName[cls.__name__.upper()]

    @staticmethod
    def class_(card_name: str) -> type[Card]:
        return Card.types.get(CardName[card_name.upper()], Card)

    def __init_subclass__(cls) -> None:
        Card.types[cls.card_name()] = cls

    def __eq__(self, other: object) -> bool:
        return isinstance(self, type(other)) or self.__class__ is other

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return self.name

    @classmethod
    def action(cls, player: Player) -> None:
        for _ in range(cls.more_cards_from_deck):
            card_name = player.take_one_card_from_deck()
            if card_name is not None:
                player.hand.append(card_name)
        player.purchases_left += cls.more_purchases
        player.actions_left += cls.more_actions
        player.money += cls.more_money

        with ErrorManager(player):
            cls._action(player)

    @classmethod
    def _action(cls, player: Player) -> None:
        pass

    @classmethod
    def buy(cls, player: Player) -> None:
        with ErrorManager(player):
            cls._buy(player)

    @classmethod
    def _buy(cls, player: Player) -> None:
        pass


class Adventurer(Card):
    name = "Aventurier"
    cost = 6
    is_action = True
    nb_treasure_cards = 2

    @classmethod
    def _action(cls, player: Player) -> None:
        nb_kept_cards = 0
        while nb_kept_cards < cls.nb_treasure_cards:
            card_name = player.take_one_card_from_deck()
            if card_name is None:
                break
            if Card.class_(card_name).is_treasure:
                nb_kept_cards += 1
                player.hand.append(card_name)
            else:
                player.discard.append(card_name)


class Artificer(Card):
    name = "Maître artisan"
    card_set = "adventures"
    cost = 5
    is_action = True
    more_actions = 1
    more_cards_from_deck = 1
    more_money = 1

    @classmethod
    def _action(cls, player: Player) -> None:
        nb_discarded_cards = 0
        for card_name in player.hand.copy():
            if player.use_hook(
                player.hooks.confirm_discard_card_from_hand,
                CardNameAndHand(card_name=card_name, hand=list(player.hand)),
            ):
                player.hand.remove(card_name)
                player.discard.append(card_name)
                nb_discarded_cards += 1

        possible_cards: list[CardName] = list({
            card_name
            for card_name in player.game.stock
            if player.game.stock.quantity(card_name)
            and Card.class_(card_name).cost == nb_discarded_cards
        })
        if possible_cards:
            chosen_card_name = player.use_hook(
                player.hooks.choose_card_to_receive_in_deck,
                PossibleCards(possible_cards=possible_cards),
            )
            card_name = CardName[chosen_card_name.upper()]
            player.game.stock.remove(card_name)
            player.deck.prepend(card_name)


class Bandit(Card):
    name = "Bandit"
    card_set = "baseset2add"
    cost = 5
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        if CardName.GOLD in player.game.stock:
            player.game.stock.remove(CardName.GOLD)
            player.discard.append(CardName.GOLD)
        for other_player in player.other_players():
            with ErrorManager(other_player):
                drawn_cards = [
                    card_name
                    for _ in range(2)
                    if (card_name := other_player.take_one_card_from_deck()) is not None
                ]
                trashable_money_cards = [
                    card_name
                    for card_name in drawn_cards
                    if card_name in (treasure_card_name - {CardName.COPPER})
                ]
                trashable_money_cards.sort(key=lambda c: Card.class_(c).money)
                if trashable_money_cards:
                    drawn_cards.remove(trashable_money_cards[0])
                for card_name in drawn_cards:
                    other_player.discard.append(card_name)


class Bureaucrat(Card):
    name = "Bureaucrate"
    cost = 4
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        silver = Silver.card_name()
        if silver in player.game.stock:
            player.deck.prepend(silver)
            player.game.stock.remove(silver)
        for other_player in player.other_players():
            with ErrorManager(other_player):
                victory_cards = other_player.hand.victory_cards
                if victory_cards:
                    other_player.deck.prepend(victory_cards[0])
                    other_player.hand.remove(victory_cards[0])


class Cellar(Card):
    name = "Cave"
    cost = 2
    is_action = True
    more_actions = 1

    @classmethod
    def _action(cls, player: Player) -> None:
        nb_discarded_cards = 0
        for card_name in player.hand.copy():
            if player.use_hook(
                player.hooks.confirm_discard_card_from_hand,
                CardNameAndHand(card_name=card_name, hand=list(player.hand)),
            ):
                player.hand.remove(card_name)
                player.discard.append(card_name)
                nb_discarded_cards += 1
        for _ in range(nb_discarded_cards):
            discarded_card_name = player.take_one_card_from_deck()
            if discarded_card_name is not None:
                player.hand.append(discarded_card_name)


class Chancellor(Card):
    name = "Chancelier"
    cost = 3
    is_action = True
    more_money = 2

    @classmethod
    def _action(cls, player: Player) -> None:
        if player.use_hook(player.hooks.confirm_discard_deck):
            player.deck.empty_to(player.discard)


class Chapel(Card):
    name = "Chapelle"
    cost = 2
    is_action = True
    max_trashed_cards: Final[int] = 4

    @classmethod
    def _action(cls, player: Player) -> None:
        nb_trashed_cards = 0
        for card_name in player.hand.copy():
            if player.use_hook(
                player.hooks.confirm_trash_card_from_hand,
                CardNameAndHand(card_name=card_name, hand=list(player.hand)),
            ):
                player.hand.remove(card_name)
                nb_trashed_cards += 1
            if nb_trashed_cards >= cls.max_trashed_cards:
                break


class Colony(Card):
    name = "Colonie"
    card_set = "prosperity"
    cost = 11
    is_kingdom = False
    is_victory = True
    victory_points = 10


class Copper(Card):
    name = "Cuivre"
    cost = 0
    money = 1
    is_kingdom = False
    is_treasure = True


class CouncilRoom(Card):
    name = "Chambre du conseil"
    cost = 5
    is_action = True
    more_cards_from_deck = 4
    more_purchases = 1

    @classmethod
    def _action(cls, player: Player) -> None:
        for other_player in player.other_players():
            with ErrorManager(other_player):
                card_name = other_player.take_one_card_from_deck()
                if card_name is not None:
                    other_player.hand.append(card_name)


class Curse(Card):
    name = "Malédiction"
    cost = 0
    is_kingdom = False
    is_victory = True
    victory_points = -1


class CursedGold(Card):
    name = "Or maudit"
    card_set = "nocturne"
    cost = 4
    money = 3
    is_kingdom = False
    is_treasure = True

    @classmethod
    def _buy(cls, player: Player) -> None:
        if CardName.CURSE in player.game.stock:
            player.game.stock.remove(CardName.CURSE)
            player.discard.append(CardName.CURSE)


class DistantShore(Card):
    name = "Rivage lointain"
    card_set = "allies"
    cost = 6
    is_action = True
    is_victory = True
    more_actions = 1
    more_cards_from_deck = 2
    victory_points = 2

    @classmethod
    def _action(cls, player: Player) -> None:
        if CardName.ESTATE in player.game.stock:
            player.game.stock.remove(CardName.ESTATE)
            player.discard.append(CardName.ESTATE)


class Duchy(Card):
    name = "Duché"
    cost = 5
    is_kingdom = False
    is_victory = True
    victory_points = 3


class Estate(Card):
    name = "Domaine"
    cost = 2
    is_kingdom = False
    is_victory = True
    victory_points = 1


class FairGrounds(Card):
    name = "Champ de foire"
    card_set = "cornucopia"
    cost = 6
    is_victory = True


class FarmingVillage(Card):
    name = "Village agricole"
    card_set = "cornucopia"
    cost = 4
    is_action = True
    more_actions = 2

    @classmethod
    def _action(cls, player: Player) -> None:
        drawn_cards = CardContainer()
        while (card_name := player.take_one_card_from_deck()) is not None:
            class_ = Card.class_(card_name)
            if class_.is_treasure or class_.is_action:
                player.hand.append(card_name)
                break
            drawn_cards.append(card_name)
        drawn_cards.empty_to(player.discard)


class Feast(Card):
    name = "Festin"
    cost = 4
    is_action = True
    max_new_card_cost: Final[int] = 5

    @classmethod
    def _action(cls, player: Player) -> None:
        player.played_cards.pop()
        possible_cards: list[CardName] = list({
            card_name
            for card_name in player.game.stock
            if player.game.stock.quantity(card_name)
            and Card.class_(card_name).cost <= cls.max_new_card_cost
        })
        if possible_cards:
            chosen_card_name = player.use_hook(
                player.hooks.choose_card_to_receive_in_discard,
                PossibleCards(possible_cards=possible_cards),
            )
            card_name = CardName[chosen_card_name.upper()]
            player.game.stock.remove(card_name)
            player.discard.append(card_name)


class Festival(Card):
    name = "Festival"
    cost = 5
    is_action = True
    more_purchases = 1
    more_actions = 2
    more_money = 2


class FortuneTeller(Card):
    name = "Diseuse de bonne aventure"
    card_set = "cornucopia"
    cost = 3
    is_action = True
    more_money = 2

    @classmethod
    def _action(cls, player: Player) -> None:
        for other_player in player.other_players():
            with ErrorManager(other_player):
                drawn_cards = CardContainer()
                while (card_name := other_player.take_one_card_from_deck()) is not None:
                    if Card.class_(card_name).is_victory:
                        other_player.deck.prepend(card_name)
                        break
                    drawn_cards.append(card_name)
                drawn_cards.empty_to(other_player.discard)


class Gardens(Card):
    name = "Jardins"
    cost = 4
    is_victory = True


class Gold(Card):
    name = "Or"
    cost = 6
    money = 3
    is_kingdom = False
    is_treasure = True


class Harvest(Card):
    name = "Récolte"
    card_set = "cornucopia"
    cost = 5
    is_action = True
    nb_drawn_cards: Final[int] = 4

    @classmethod
    def _action(cls, player: Player) -> None:
        drawn_cards = CardContainer()
        while (card_name := player.take_one_card_from_deck()) is not None:
            drawn_cards.append(card_name)
            if len(drawn_cards) == cls.nb_drawn_cards:
                break
        nb_different_cards = len(set(drawn_cards))
        drawn_cards.empty_to(player.discard)
        for _ in range(nb_different_cards):
            card_name = player.take_one_card_from_deck()
            if card_name is None:
                break
            player.hand.append(card_name)


class Hireling(Card):
    name = "Recrue"
    card_set = "adventures"
    cost = 6
    is_action = True
    more_cards_from_deck = 1

    @classmethod
    def _action(cls, player: Player) -> None:
        player.played_cards.pop()
        player.nb_cards_in_hand_at_turn_start += 1


class Laboratory(Card):
    name = "Laboratoire"
    cost = 5
    is_action = True
    more_cards_from_deck = 2
    more_actions = 1


class Library(Card):
    name = "Bibliothèque"
    cost = 5
    is_action = True
    target_hand_size: Final[int] = 7

    @classmethod
    def _action(cls, player: Player) -> None:
        skipped_cards = CardContainer()
        while len(player.hand) < cls.target_hand_size:
            card_name = player.take_one_card_from_deck()
            if card_name is None:
                break
            if not Card.class_(card_name).is_action:
                player.hand.append(card_name)
            elif player.use_hook(
                player.hooks.skip_card_reception_in_hand,
                CardNameAndHand(card_name=card_name, hand=list(player.hand)),
            ):
                skipped_cards.append(card_name)
            else:
                player.hand.append(card_name)
        skipped_cards.empty_to(player.discard)


class Magnate(Card):
    name = "Magnat"
    card_set = "prosperity2add"
    cost = 5
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        for _ in list(player.hand.money_cards):
            card_name = player.take_one_card_from_deck()
            if card_name is not None:
                player.hand.append(card_name)


class Magpie(Card):
    name = "Pie voleuse"
    card_set = "adventures"
    cost = 4
    is_action = True
    more_actions = 1
    more_cards_from_deck = 1

    @classmethod
    def _action(cls, player: Player) -> None:
        card_name = player.take_one_card_from_deck()
        if card_name is not None:
            if card_name in treasure_card_name:
                player.hand.append(card_name)
            else:
                player.deck.prepend(card_name)
                if CardName.MAGPIE in player.game.stock:
                    player.game.stock.remove(CardName.MAGPIE)
                    player.discard.append(CardName.MAGPIE)


class Market(Card):
    name = "Marché"
    cost = 5
    is_action = True
    more_cards_from_deck = 1
    more_purchases = 1
    more_actions = 1
    more_money = 1


class Marquis(Card):
    name = "Marquis"
    card_set = "allies"
    cost = 6
    is_action = True
    more_purchases = 1
    hand_size_left: Final[int] = 10

    @classmethod
    def _action(cls, player: Player) -> None:
        for _ in list(player.hand):
            card_name = player.take_one_card_from_deck()
            if card_name is not None:
                player.hand.append(card_name)
        while len(player.hand) > cls.hand_size_left:
            removed_card = player.use_hook(
                player.hooks.discard_card_from_hand,
                Hand(hand=list(player.hand)),
            )
            player.hand.remove(CardName[removed_card.upper()])
            player.discard.append(CardName[removed_card.upper()])


class Militia(Card):
    name = "Milice"
    cost = 4
    is_action = True
    more_money = 2
    enemy_hand_size_left: Final[int] = 3

    @classmethod
    def _action(cls, player: Player) -> None:
        for other_player in player.other_players():
            with ErrorManager(other_player):
                while len(other_player.hand) > cls.enemy_hand_size_left:
                    removed_card = other_player.use_hook(
                        other_player.hooks.discard_card_from_hand,
                        Hand(hand=list(other_player.hand)),
                    )
                    other_player.hand.remove(CardName[removed_card.upper()])
                    other_player.discard.append(CardName[removed_card.upper()])


class Mine(Card):
    name = "Mine"
    cost = 5
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        money_cards: list[str] = list(set(player.hand.money_cards))
        if money_cards:
            trashed_card = player.use_hook(
                player.hooks.trash_money_card_for_better_money_card,
                MoneyCardsInHand(money_in_hand=money_cards),
            )
            if trashed_card is not None:
                trashed_card_name = CardName[trashed_card.upper()]
                player.hand.remove(trashed_card_name)
                possible_money_cards = [
                    card_name
                    for card_name in player.game.stock
                    if (class_ := Card.class_(card_name)).is_treasure
                    and class_.cost <= Card.class_(trashed_card_name).cost + 3
                    and player.game.stock.quantity(card_name)
                ]
                if possible_money_cards:
                    best_money = max(
                        possible_money_cards,
                        key=lambda card_name: Card.class_(card_name).money,
                    )
                    player.hand.append(best_money)


class MoneyLender(Card):
    name = "Prêteur sur gages"
    cost = 4
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        if player.hand.copper_qty >= 1 and player.use_hook(
            player.hooks.confirm_trash_card_from_hand,
            CardNameAndHand(card_name=CardName.COPPER, hand=list(player.hand)),
        ):
            player.hand.remove(CardName.COPPER)
            player.money += 3


class Platinum(Card):
    name = "Platine"
    card_set = "prosperity"
    cost = 9
    money = 5
    is_kingdom = False
    is_treasure = True


class Poacher(Card):
    name = "Braconnier"
    card_set = "baseset2add"
    cost = 4
    is_action = True
    more_cards_from_deck = 1
    more_actions = 1
    more_money = 1

    @classmethod
    def _action(cls, player: Player) -> None:
        for _ in range(player.game.stock.nb_empty_piles):
            if not player.hand:
                break
            removed_card = player.use_hook(
                player.hooks.discard_card_from_hand,
                Hand(hand=list(player.hand)),
            )
            player.hand.remove(CardName[removed_card.upper()])
            player.discard.append(CardName[removed_card.upper()])


class Port(Card):
    name = "Ville portuaire"
    card_set = "adventures"
    cost = 4
    is_action = True
    more_cards_from_deck = 1
    more_actions = 2


class Province(Card):
    name = "Province"
    cost = 8
    is_kingdom = False
    is_victory = True
    victory_points = 6


class Remake(Card):
    name = "Renouvellement"
    card_set = "cornucopia"
    cost = 4
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        for _ in range(2):
            if not player.hand:
                return
            trashed_card = player.use_hook(
                player.hooks.trash_card_from_hand,
                Hand(hand=list(player.hand)),
            )
            trashed_card_name = CardName[trashed_card.upper()]
            player.hand.remove(trashed_card_name)
            possible_cards: list[CardName] = list({
                card_name
                for card_name in player.game.stock
                if player.game.stock.quantity(card_name)
                and Card.class_(card_name).cost
                == Card.class_(trashed_card_name).cost + 1
            })
            if possible_cards:
                chosen_card = player.use_hook(
                    player.hooks.choose_card_to_receive_in_discard,
                    PossibleCards(possible_cards=possible_cards),
                )
                chosen_card_name = CardName[chosen_card.upper()]
                player.game.stock.remove(chosen_card_name)
                player.discard.append(chosen_card_name)


class Remodel(Card):
    name = "Rénovation"
    cost = 4
    is_action = True

    @classmethod
    def _action(cls, player: Player) -> None:
        if not player.hand:
            return
        trashed_card = player.use_hook(
            player.hooks.trash_card_from_hand,
            Hand(hand=list(player.hand)),
        )
        trashed_card_name = CardName[trashed_card.upper()]
        player.hand.remove(trashed_card_name)
        possible_cards: list[CardName] = list({
            card_name
            for card_name in player.game.stock
            if player.game.stock.quantity(card_name)
            and Card.class_(card_name).cost <= Card.class_(trashed_card_name).cost + 2
        })
        if possible_cards:
            chosen_card = player.use_hook(
                player.hooks.choose_card_to_receive_in_discard,
                # TODO partout où il y a des possible_cards, vérifier que la réponse est
                # dans la liste
                PossibleCards(possible_cards=possible_cards),
            )
            chosen_card_name = CardName[chosen_card.upper()]
            player.game.stock.remove(chosen_card_name)
            player.discard.append(chosen_card_name)


class Silver(Card):
    name = "Argent"
    cost = 3
    money = 2
    is_kingdom = False
    is_treasure = True


class Smithy(Card):
    name = "Forgeron"
    cost = 4
    is_action = True
    more_cards_from_deck = 3


class Swap(Card):
    name = "Troc"
    card_set = "allies"
    cost = 5
    is_action = True
    more_cards_from_deck = 1
    more_actions = 1
    max_cost_swap = 5

    @classmethod
    def _action(cls, player: Player) -> None:
        for trashed_card_name in list(player.hand):
            if not Card.class_(trashed_card_name).is_action:
                continue
            # TODO c'est pas vraiment un trash, il faudrait utiliser un nouveau hook
            if player.use_hook(
                player.hooks.confirm_trash_card_from_hand,
                CardNameAndHand(card_name=trashed_card_name, hand=list(player.hand)),
            ):
                player.hand.remove(trashed_card_name)
                player.game.stock.append(trashed_card_name)

                possible_cards: list[CardName] = list({
                    card_name
                    for card_name in player.game.stock
                    if player.game.stock.quantity(card_name)
                    and Card.class_(card_name).cost <= cls.max_cost_swap
                    and Card.class_(card_name).is_action
                    and card_name != trashed_card_name
                })
                if possible_cards:
                    # TODO normalement ça ne va pas en discard, mais dans la main,
                    # nouveau hook à prévoir
                    chosen_card = player.use_hook(
                        player.hooks.choose_card_to_receive_in_discard,
                        # TODO partout où il y a des possible_cards, vérifier que la
                        # réponse est dans la liste
                        PossibleCards(possible_cards=possible_cards),
                    )
                    chosen_card_name = CardName[chosen_card.upper()]
                    player.game.stock.remove(chosen_card_name)
                    player.discard.append(chosen_card_name)

                break


class Village(Card):
    name = "Village"
    cost = 3
    is_action = True
    more_cards_from_deck = 1
    more_actions = 2


class Witch(Card):
    name = "Sorcière"
    cost = 5
    is_action = True
    more_cards_from_deck = 2

    @classmethod
    def _action(cls, player: Player) -> None:
        for other_player in player.other_players():
            with ErrorManager(other_player):
                if player.game.stock.curse_qty:
                    other_player.discard.append(CardName.CURSE)
                    player.game.stock.remove(CardName.CURSE)


class Woodcutter(Card):
    name = "Bucheron"
    cost = 3
    is_action = True
    more_purchases = 1
    more_money = 2


class Workshop(Card):
    name = "Atelier"
    cost = 3
    is_action = True
    max_cost_of_received_card: Final[int] = 4

    @classmethod
    def _action(cls, player: Player) -> None:
        possible_cards: list[str] = list({
            card_name
            for card_name in player.game.stock
            if player.game.stock.quantity(card_name)
            and Card.class_(card_name).cost <= cls.max_cost_of_received_card
        })
        possible_cards = list(set(possible_cards))
        logger.debug(possible_cards)
        if possible_cards:
            chosen_card = player.use_hook(
                player.hooks.choose_card_to_receive_in_discard,
                PossibleCards(possible_cards=possible_cards),
            )
            chosen_card_name = CardName[chosen_card.upper()]
            player.game.stock.remove(chosen_card_name)
            player.discard.append(chosen_card_name)


actions_card_name: set[CardName] = {
    CardName[name.upper()]
    for name, class_ in inspect.getmembers(sys.modules[__name__], inspect.isclass)
    if issubclass(class_, Card) and class_.name != "Unknown" and class_.is_action
}

treasure_card_name: set[CardName] = {
    CardName[name.upper()]
    for name, class_ in inspect.getmembers(sys.modules[__name__], inspect.isclass)
    if issubclass(class_, Card) and class_.name != "Unknown" and class_.is_treasure
}

victory_card_name: set[CardName] = {
    CardName[name.upper()]
    for name, class_ in inspect.getmembers(sys.modules[__name__], inspect.isclass)
    if issubclass(class_, Card) and class_.name != "Unknown" and class_.is_victory
}


class CardContainer:
    """Storage for card."""

    def __init__(self) -> None:
        self._quantities: dict[CardName, int] = defaultdict(int)
        self._cards: list[CardName] = []

    def __getattr__(self, attribute: str) -> int:
        if attribute.endswith("_qty"):
            card_name = attribute[:-4]
            return self._quantities.get(CardName[card_name.upper()], 0)
        raise AttributeError

    def __repr__(self) -> str:
        return str(self._cards)

    def __add__(self, other: CardContainer) -> CardContainer:
        ret = CardContainer()
        for card_name in self:
            ret.append(card_name)
        for card_name in other:
            ret.append(card_name)
        return ret

    def __contains__(self, card_name: CardName) -> bool:
        return self._quantities.get(card_name, 0) > 0

    def __len__(self) -> int:
        return len(self._cards)

    def __getitem__(self, index: int) -> CardName:
        return self._cards[index]

    def __iter__(self) -> Iterator[CardName]:
        return iter(self._cards)

    def clear(self) -> None:
        self._quantities.clear()
        self._cards.clear()

    def copy(self) -> CardContainer:
        new = CardContainer()
        new._quantities = self._quantities.copy()
        new._cards = self._cards.copy()
        return new

    def prepend(self, card_name: CardName) -> None:
        """
        Insert next card to be played.

        Raises:
            KeyError: If unknown card_name.

        """
        if card_name not in Card.types:
            raise KeyError(card_name)
        self._cards.insert(0, card_name)
        self._quantities[card_name] += 1

    def append(self, card_name: CardName) -> None:
        if card_name not in Card.types:
            raise KeyError(card_name)
        self._cards.append(card_name)
        self._quantities[card_name] += 1

    def append_several(self, qty: int, card_name: CardName) -> None:
        for _ in range(qty):
            self.append(card_name)

    def remove(self, card_name: CardName) -> None:
        self._cards.remove(card_name)
        self._quantities[card_name] -= 1

    def shuffle(self) -> None:
        random.shuffle(self._cards)

    def contains_action(self) -> bool:
        return any(
            card_name in actions_card_name
            for card_name, qty in self._quantities.items()
            if qty > 0
        )

    @property
    def victory_cards(self) -> CardContainer:
        ret = CardContainer()
        for card_name, qty in self._quantities.items():
            if card_name in victory_card_name:
                ret.append_several(qty, card_name)
        return ret

    @property
    def action_cards(self) -> CardContainer:
        ret = CardContainer()
        for card_name, qty in self._quantities.items():
            if card_name in actions_card_name:
                ret.append_several(qty, card_name)
        return ret

    def contains_money(self) -> bool:
        return any(
            card_name in treasure_card_name
            for card_name, qty in self._quantities.items()
            if qty > 0
        )

    @property
    def money_cards(self) -> CardContainer:
        ret = CardContainer()
        for card_name, qty in self._quantities.items():
            if card_name in treasure_card_name:
                ret.append_several(qty, card_name)
        return ret

    @property
    def money(self) -> int:
        return sum(
            Card.class_(card_name).money * quantity
            for card_name, quantity in self._quantities.items()
        )

    def pop(self, index: int = -1) -> CardName:
        card_name = self._cards.pop(index)
        self._quantities[card_name] -= 1
        return card_name

    def quantity(self, card_name: CardName) -> int:
        return self._quantities.get(card_name, 0)

    def sort(self, key: Callable, *, reverse: bool = False) -> None:
        self._cards.sort(key=key, reverse=reverse)

    def empty_to(self, other: CardContainer) -> None:
        while self:
            card_name = self.pop(0)
            other.append(card_name)
        self._quantities.clear()
        self._cards.clear()

    @property
    def nb_empty_piles(self) -> int:
        return sum(1 for qty in self._quantities.values() if qty == 0)

    @property
    def three_empty_piles(self) -> bool:
        nb_piles = 3
        return self.nb_empty_piles >= nb_piles

    @property
    def state(self) -> Cards:
        return Cards(
            quantities={
                card_name: quantity
                for card_name, quantity in self._quantities.items()
                if quantity > 0
            },
        )
